import sys,os,types

import admit.util.bdp_types as ut
from admit.bdp import *
from admit.at import *
from admit.util.Table import Table
from admit.util.Image import Image
import admit.util.registry as registry

"""
This module is used to generate the following:
      all dtd files   xml reading/writing definition files
      types.py     Constants and types definition file

Usage:
     import dtdGenerator
     dtdGenerator.generate()
     
Relies on:
    registry.py    this file contains a listing of all current BDP classes
"""

# Just some useful constant strings
START = "<!ELEMENT "
END = "(#PCDATA)>\n"
ATSTART = "<!ATTLIST "
ATEND = ") #REQUIRED>\n"
HEADER = "<!-- The root node is BDP and has child nodes listed afterward,\n     a * indicates 0 or more instances\n     a ? indicates 0 or 1 instances\n     no added marker indicates 1 instance is expected\n -->\n"
MIDHEAD = "<!-- attributes for the different nodes\n     format: node attribute (choices if like an enum)\n     CDATA indicates character data, followed by the default value\n-->\n"
AUTOGEN = "This file is automatically generated by dtdGenerator.py,\n  based on entries in registry.py. Edit this file at your own risk."
ALLTYPES = " type (INT|FLOAT|DICT|LIST|BOOL|LONG|STRING|NDARRAY|TUPLE) #REQUIRED>\n"
NDARR = " ndarray (1|0) #REQUIRED>\n"
atlist = []
T = Table()
I = Image()


def writeImage(name) :
    """
    Method to write out the image data type to a dtd
    Input: name - the name of the image node
    """
    out = START + name + "\t(fileName,xmin,xmax,ymin,ymax,description)>\n"
    out += ATSTART + name + " type (" + ut.IMAGE + ATEND
    out += START + "fileName\t\t" + END
    out += ATSTART + "fileName type (" + ut.STRING + ATEND
    out += START + "xmin\t\t" + END
    out += ATSTART + "xmin type (" + ut.FLOAT + ATEND
    out += START + "xmax\t\t" + END
    out += ATSTART + "xmax type (" + ut.FLOAT + ATEND
    out += START + "ymin\t\t" + END
    out += ATSTART + "ymin type (" + ut.FLOAT + ATEND
    out += START + "ymax\t\t" + END
    out += ATSTART + "ymax type (" + ut.FLOAT + ATEND
    out += START + "description\t\t" + END
    out += ATSTART + "description type (" + ut.STRING + ATEND
    return out

def writeTable(name) :
    """
    Method to write out the table data type to a dtd
    Input: name - the name of the table node
    """
    out = START + name + "\t(columns,units,planes,data,min,max)>\n"
    out += ATSTART + name + " type (" + ut.TABLE + ATEND
    out += START + "columns\t\t" + END
    out += ATSTART + "columns type (" + ut.LIST + ATEND
    out += ATSTART + "columns " + NDARR
    out += START + "units\t\t" + END
    out += ATSTART + "units type (" + ut.LIST + ATEND
    out += ATSTART + "units " + NDARR
    out += START + "planes\t\t" + END
    out += ATSTART + "planes type (" + ut.LIST + ATEND
    out += ATSTART + "planes" + NDARR
    out += START + "data\t\t" + END
    out += ATSTART + "data type (" + ut.NDARRAY + ATEND
    out += START + "min\t\t" + END
    out += ATSTART + "min type (" + ut.LIST + ATEND
    out += ATSTART + "min " + NDARR
    out += START + "max\t\t" + END
    out += ATSTART + "max type (" + ut.LIST + ATEND
    out += ATSTART + "max " + NDARR
    return out

def writeTask(name) :
    """
    Method to write out the ADMIT task data type to a dtd
    Input: name - the name of the task node
    """
    out = START + name + "\t(taskName,keywords,version)>\n"
    out += ATSTART + name + " type (" + ut.TASK + ATEND
    out += START + "taskName\t\t" + END
    out += ATSTART + "taskName type (" + ut.STRING + ATEND
    out += START + "keywords\t\t" + END
    out += ATSTART + "keywords type (" + ut.DICT + ATEND
    out += ATSTART + "keywords " + NDARR
    out += START + "version\t\t" + END
    out += ATSTART + "version type (" + ut.STRING + ATEND
    return out

def write(key,val):
    """
    Method which takes a class name and BDP sub-class instance. It creates
    a dtd file for the class and collects relevant info for the ut
    file.
    Input: key - the name of the class
           value - an instance of the BDP sub-class
    """
    sys.stdout.write("Processing %s ..." % (key))
    sys.stdout.flush()
    # open the dtd file, we don't append, just write it from scratch
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + os.sep + "..")

    dtd = open(os.path.dirname(os.path.realpath(__file__)) + os.sep + "dtd" + os.sep + key + ".dtd",'w')
    dtd.write("<!-- " + AUTOGEN + "\n-->\n")
    dtd.write(HEADER)
    tables = []
    images = []
    tasks = []
    ndar = []
    listing = []
    # write out the individual data elements based on their occurrence
    line = START + ut.BDP + "\t("
    # loop over all data nodes adding them to the appropriate list
    t = val
    for i in t.__dict__:
        if(type(getattr(t,i)) == type(T)) :
            tables.append(i)
        elif(type(getattr(t,i)) == type(I)) :
            images.append(i)
        elif(type(getattr(t,i)) == types.ListType or type(getattr(t,i)) == types.DictionaryType) :
            ndar.append(i)
        listing.append(i)
        line += i + ","
    line = line[:-1]
    line += ")>\n"
    dtd.write(line)
    dtd.write(ATSTART + "BDP type (" + key + ") #REQUIRED>\n")
    # process images, tables, tasks, and everything else adding their attributes
    dtd.write("<!-- child nodes #PCDATA indicates parsable character data-->\n")
    for j in listing:
        if(j in tables) :
            dtd.write(writeTable(j))
        elif(j in images) :
            dtd.write(writeImage(j))
        else :
            dtd.write(START + j + "\t\t" + END)
            dtd.write(ATSTART + j + ALLTYPES)
            if(j in ndar) :
                dtd.write(ATSTART + j + NDARR)

    dtd.close()
    sys.stdout.write("Done\n")
    sys.stdout.flush()

def generate():

    # process all other classes, collecting their names
    for key in registry.BDPTYPES :
        write(key,registry.BDPTYPES[key])

    ATTYPES = ""
    for key in registry.ATTYPES :
        ATTYPES += key.upper() + "|"
    ATTYPES = ATTYPES[:-1]
    admit = open(os.path.dirname(os.path.realpath(__file__)) + os.sep + "dtd" + os.sep + "admit.dtd",'w')
    admit.write("<!-- " + AUTOGEN + "\n-->\n")
    admit.write(HEADER)
    admit.write(START + "ADMIT (userData,AT*)>\n")
    admit.write(START + "userData " + END)
    admit.write(ATSTART + "userData type (DICT" + ATEND)
    admit.write(ATSTART + "userData " + NDARR)
    admit.write(START + "AT " + END)
    admit.write(ATSTART + "AT type (" + ATTYPES + ATEND)
    admit.write(ATSTART + "AT id (STRING|ANY" + ATEND)
    admit.write(ATSTART + "AT version (STRING|ANY" + ATEND)
    admit.write(ATSTART + "AT parameters (DICT|ANY" + ATEND)
    admit.write(ATSTART + "AT stale (0|1" + ATEND)
    admit.close()

    # create the ut.py file
    types = open(os.path.dirname(os.path.realpath(__file__)) + os.sep + ".." + os.sep +"util" + os.sep + "bdp_types.py",'w')
    types.write("\"\"\"\n" + AUTOGEN + "\n\nThis file contains string constants useful for programming\n\"\"\"\n")
    sys.stdout.write("Writing types.py ...")
    sys.stdout.flush()
    # these are really just constant strings that will make life easier for programming
    for key in registry.BDPTYPES :
        types.write(key.upper() + " = \"" + key + "\"\n")
    for key in registry.ATTYPES :
        types.write(key.upper() + " = \"" + key + "\"\n")
    types.write("GENERIC = \"Generic\"\n")
    types.write("ADMIT = \"ADMIT\"\n")
    types.write("AT = \"AT\"\n")
    types.write("BDP = \"BDP\"\n")
    types.write("INT = \"INT\"\n")
    types.write("FLOAT = \"FLOAT\"\n")
    types.write("DICT = \"DICT\"\n")
    types.write("LIST = \"LIST\"\n")
    types.write("BOOL = \"BOOL\"\n")
    types.write("LONG = \"LONG\"\n")
    types.write("STRING = \"STRING\"\n")
    types.write("NDARRAY = \"NDARRAY\"\n")
    types.write("TUPLE = \"TUPLE\"\n")
    types.write("IMAGE = \"IMAGE\"\n")
    types.write("TABLE = \"TABLE\"\n")
    types.write("TASK = \"TASK\"\n")
    types.close()
    
    sys.stdout.write("Done\n")
    sys.stdout.flush()
    
